
# This file is automatically generated by the setup.py script
long_description = """
JSONSchema API Generator
========================
The schemaperfect_ package provides tools for auto-generation of Python
APIs from JSONSchema_ specifications.

See more information in the README_.

.. _README: https://github.com/jwilson8767/schemaperfect/blob/master/README.md
.. _JSONSchema: http://json-schema.org/
.. _schemaperfect: https://github.com/jwilson8767/schemaperfect
"""
short_version = '0.5.0'
version = '0.5.0'
full_version = '0.5.0'
git_revision = 'ffc2ca9ded10bc31651e6867d630b0d3d5daf8d9'
release = True
if not release:
    version = full_version
